#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pyn5` package."""
import unittest
from pathlib import Path
import shutil
import json

import numpy as np
import pytest

import pyn5

from .common import blocks_in, attrs_in, z5py, blocks_hash
from .conftest import BLOCKSIZE


def dtype_info(dtype):
    if np.issubdtype(dtype, np.integer):
        info = np.iinfo(dtype)
    elif np.issubdtype(dtype, np.floating):
        info = np.finfo(dtype)
    else:
        raise ValueError("Unknown dtype {}".format(dtype))

    return info


def valid_block(dtype):
    info = dtype_info(dtype)
    return np.linspace(info.min, info.max, np.product(BLOCKSIZE), dtype=dtype).tolist()


def overflow_block(dtype):
    info = dtype_info(dtype)
    length = np.product(BLOCKSIZE) // 2
    before = [info.min - i - 1 for i in range(length)]
    after = [info.max + i + 1 for i in range(length)]
    return before + after


def wrong_dtype_block(dtype):
    if np.issubdtype(dtype, np.integer):
        return np.linspace(0, 1, np.product(BLOCKSIZE)).tolist()
    else:
        return np.asarray(valid_block(dtype), dtype=np.dtype("int64")).tolist()


def ravel_block(block):
    return np.array(block).reshape(BLOCKSIZE, order="F")


def test_read_write_valid(ds_dtype):
    ds, dtype = ds_dtype
    block_idx = [0, 0, 0]

    expected = valid_block(dtype)
    ds.write_block(block_idx, expected)

    np.testing.assert_equal(
        ds.read_ndarray(block_idx, BLOCKSIZE), ravel_block(expected)
    )


def test_read_write_overflow(ds_dtype):
    ds, dtype = ds_dtype
    if np.issubdtype(dtype, np.floating):
        pytest.xfail(
            "Doesn't work properly for float types "
            "since floats that are too large get converted to inf"
        )

    block = overflow_block(dtype)

    with pytest.raises(OverflowError):
        ds.write_block([1, 1, 1], block)

    np.testing.assert_equal(ds.read_ndarray([2, 2, 2], BLOCKSIZE), np.zeros(BLOCKSIZE))


def test_read_write_wrong_dtype(ds_dtype):
    ds, dtype = ds_dtype
    if np.issubdtype(dtype, np.floating):
        pytest.xfail(
            "Doesn't work properly for float types "
            "since ints in lists don't hold their type"
        )

    block = wrong_dtype_block(dtype)

    with pytest.raises(TypeError):
        ds.write_block([2, 2, 2], block)

    np.testing.assert_equal(ds.read_ndarray([4, 4, 4], BLOCKSIZE), np.zeros(BLOCKSIZE))


def test_compression(tmp_path, compression_dict):
    root = tmp_path / "test.n5"

    data = np.arange(100, dtype=np.uint8).reshape((10, 10))

    pyn5.create_dataset(
        str(root), "ds", data.shape, (5, 5), data.dtype.name.upper(),
        json.dumps(compression_dict)
    )

    with open(root / "ds" / "attributes.json") as f:
        attrs = json.load(f)

    assert attrs["compression"] == compression_dict

    ds = pyn5.DatasetUINT8(str(root), "ds", True)
    ds.write_ndarray((0, 0), data, 0)


def test_default_compression(tmp_path):
    root = tmp_path / "test.n5"

    data = np.arange(100, dtype=np.uint8).reshape((10, 10))

    pyn5.create_dataset(
        str(root), "ds", data.shape, (5, 5), data.dtype.name.upper(),
    )

    with open(root / "ds" / "attributes.json") as f:
        attrs = json.load(f)

    assert attrs["compression"] == {"type": "gzip", "level": -1}

    ds = pyn5.DatasetUINT8(str(root), "ds", True)
    ds.write_ndarray((0, 0), data, 0)


def test_default_compression_opts(tmp_path, compression_name_opt):
    name, _ = compression_name_opt
    root = tmp_path / "test.n5"

    data = np.arange(100, dtype=np.uint8).reshape((10, 10))

    pyn5.create_dataset(
        str(root), "ds", data.shape, (5, 5), data.dtype.name.upper(),
        json.dumps({"type": name})
    )

    with open(root / "ds" / "attributes.json") as f:
        attrs = json.load(f)

    assert attrs["compression"]["type"] == name

    ds = pyn5.DatasetUINT8(str(root), "ds", True)
    ds.write_ndarray((0, 0), data, 0)


def test_data_ordering(tmp_path):
    root = tmp_path / "test.n5"

    shape = (10, 20)
    chunks = (10, 10)

    pyn5.create_dataset(str(root), "ds", shape, chunks, "UINT8")
    ds = pyn5.DatasetUINT8(str(root), "ds", False)
    arr = np.array(ds.read_ndarray((0, 0), shape))
    arr += 1
    ds.write_ndarray((0, 0), arr, 0)

    ds_path = root / "ds"

    assert blocks_in(ds_path) == {"0", "0/0", "0/1"}

    with open(ds_path / "attributes.json") as f:
        attrs = json.load(f)

    assert list(shape) == attrs["dimensions"]


def test_vs_z5(tmp_path, z5_file):
    """Check different dimensions, same dtype/compression as z5"""
    root = tmp_path / "test.n5"

    z5_path = Path(z5_file.path)
    shape = (10, 20)
    data = np.arange(np.product(shape)).reshape(shape)
    chunks = (6, 7)

    pyn5.create_dataset(str(root), "ds", shape, chunks, data.dtype.name.upper())
    ds = pyn5.DatasetINT64(str(root), "ds", False)
    ds.write_ndarray((0, 0), data, 0)

    z5_file.create_dataset("ds", data=data, chunks=(6, 7), compression="gzip", level=-1)

    assert np.allclose(ds.read_ndarray((0, 0), shape), z5_file["ds"][:])
    assert blocks_in(root / "ds") != blocks_in(z5_path / "ds")

    attrs = attrs_in(root / "ds")
    z5_attrs = attrs_in(z5_path / "ds")
    for key in ("blockSize", "dimensions"):
        assert attrs[key] != z5_attrs[key]

    for key in ("dataType", "compression"):
        assert attrs[key] == z5_attrs[key]

    data2 = pyn5.DatasetINT64(str(z5_path), "ds", False).read_ndarray((0, 0), shape)
    data3 = z5py.N5File(root)["ds"][:]

    assert not all([
        np.array_equal(data, data2),
        np.array_equal(data, data3)
    ])


def test_vs_z5_hash(tmp_path, z5_file):
    """Check different block hashes to z5"""
    root = tmp_path / "test.n5"

    z5_path = Path(z5_file.path)
    shape = (10, 20)
    data = np.arange(np.product(shape)).reshape(shape)
    chunks = (6, 7)

    pyn5.create_dataset(
        str(root), "ds", shape, chunks, data.dtype.name.upper(), json.dumps({"type": "raw"})
    )
    ds = pyn5.DatasetINT64(str(root), "ds", False)
    ds.write_ndarray((0, 0), data, 0)

    z5_file.create_dataset("ds", data=data, chunks=(6, 7), compression="raw")

    assert blocks_hash(root) != blocks_hash(z5_path)


class TestPythonReadWrite(unittest.TestCase):
    def setUp(self):
        self.root = "test.n5"
        self.dataset = "test"
        self.dtype = "UINT8"
        self.dataset_size = [10, 10, 10]
        self.block_size = [2, 2, 2]

        pyn5.create_dataset(
            self.root, self.dataset, self.dataset_size, self.block_size, self.dtype
        )
        self.n5 = pyn5.open(self.root, self.dataset, self.dtype, False)

    def tearDown(self):
        if Path(self.root).is_dir():
            shutil.rmtree(str(Path(self.root).absolute()))

    def test_read_write(self):
        # make sure n5 is initialized to all zeros
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([0, 0, 0]), np.array(self.dataset_size))),
                np.zeros([10, 10, 10]),
            )
        )

        # write ones to whole dataset, and then
        # write on partial blocks to make sure data isn't overwritten
        pyn5.write(
            self.n5,
            (np.array([0, 0, 0]), np.array(self.dataset_size)),
            np.ones(self.dataset_size),
        )
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([0, 0, 0]), np.array(self.dataset_size))),
                np.ones([10, 10, 10]),
            )
        )

        pyn5.write(
            self.n5, (np.array([1, 1, 1]), np.array([3, 3, 3])), np.ones([2, 2, 2]) * 2
        )
        self.assertTrue(
            np.array_equal(
                pyn5.read(self.n5, (np.array([1, 1, 1]), np.array([3, 3, 3]))),
                np.ones([2, 2, 2]) * 2,
            )
        )

        # test writting non-uniform block to make sure axis orderings are correct
        self.n5.write_ndarray(
            np.array([1, 1, 1]),
            np.array(range(64), dtype=np.uint8).reshape([4, 4, 4]),
            0,
        )
        self.assertTrue(
            np.array_equal(
                self.n5.read_ndarray(np.array([1, 1, 1]), np.array([4, 4, 4])),
                np.array(range(64), dtype=int).reshape([4, 4, 4]),
            )
        )

        # sanity check axis orderings
        data = np.zeros([5, 7, 1], dtype=np.uint8)
        data[0, :, 0] = 1
        data[1, :, 0] = 2
        data[2, :, 0] = 3
        self.n5.write_ndarray(np.array([0, 0, 0]), data, 0)
        self.assertTrue(
            np.all(self.n5.read_ndarray(np.array([1, 0, 0]), np.array([1, 7, 1])) == 2)
        )
        np.testing.assert_equal(
            self.n5.read_ndarray(np.array([0, 0, 0]), np.array([3, 1, 1])).flatten(),
            np.array([1, 2, 3]),
        )

        # test writing data in non block shapes
        data = np.array(range(6), dtype=np.uint8).reshape([1, 2, 3])
        self.n5.write_ndarray(np.array([0, 0, 0]), data, 0)

        self.assertTrue(
            np.array_equal(
                self.n5.read_ndarray(np.array([0, 0, 0]), np.array([1, 2, 3])), data
            )
        )



import json

import numpy as np
import pytest
import shutil
import tempfile
from copy import deepcopy
from pathlib import Path

from h5py_like import Mode, FileMixin
from h5py_like.test_utils import (
    FileTestBase, ThreadedDatasetTestBase, GroupTestBase, ModeTestBase,
)
from pyn5 import File

from .common import blocks_hash
from .common import blocks_in, attrs_in

ds_kwargs = deepcopy(ThreadedDatasetTestBase.dataset_kwargs)
ds_kwargs["chunks"] = (5, 5, 5)


class TestFile(FileTestBase):
    dataset_kwargs = ds_kwargs
    pass


class TestGroup(GroupTestBase):
    dataset_kwargs = ds_kwargs
    pass


class TestDataset(ThreadedDatasetTestBase):
    dataset_kwargs = deepcopy(ThreadedDatasetTestBase.dataset_kwargs)

    def test_has_metadata(self, file_):
        ds = self.dataset(file_)
        with open(ds.attrs._path) as f:
            attrs = json.load(f)
        for key in ds.attrs._dataset_keys:
            assert key in attrs

    def test_no_return_metadata(self, file_):
        ds = self.dataset(file_)

        for key in ds.attrs._dataset_keys:
            assert key not in ds.attrs
            assert key not in dict(ds.attrs)

    def test_no_mutate_metadata(self, file_):
        ds = self.dataset(file_)

        for key in ds.attrs._dataset_keys:
            with pytest.raises(RuntimeError):
                ds.attrs[key] = "not a datatype"

            with pytest.raises(RuntimeError):
                del ds.attrs[key]


class TestMode(ModeTestBase):
    def setup_method(self):
        self.tmp_dir = Path(tempfile.mkdtemp())

    def teardown_method(self):
        try:
            shutil.rmtree(self.tmp_dir)
        except FileNotFoundError:
            pass

    def factory(self, mode: Mode) -> FileMixin:
        fpath = self.tmp_dir / "test.n5"
        return File(fpath, mode)


def test_data_ordering(file_, h5_file):
    shape = (5, 10, 15)
    data = np.arange(np.product(shape)).reshape(shape)
    ds_kwargs = {"shape": data.shape, "dtype": data.dtype, "chunks": (4, 4, 4)}

    for f in (file_, h5_file):
        n5_ds = f.create_dataset("ds", **ds_kwargs)
        n5_ds[:] = data

    assert np.allclose(file_["ds"][:], h5_file["ds"][:])


def test_created_dirs(file_):
    shape = (10, 20)
    data = np.ones(shape)

    ds = file_.create_dataset("ds", data=data, chunks=(10, 10))

    assert blocks_in(ds._path) == {"0", "1", "0/0", "1/0"}

    attrs = ds.attrs._read_attributes()

    assert list(ds.shape) == attrs["dimensions"][::-1]


def test_vs_z5(file_, z5_file):
    """Check same as z5"""
    z5_path = Path(z5_file.path)
    shape = (10, 20)
    data = np.arange(np.product(shape)).reshape(shape)

    file_.create_dataset("ds", data=data, chunks=(6, 7))
    z5_file.create_dataset("ds", data=data, chunks=(6, 7), compression="gzip", level=-1)

    assert np.allclose(file_["ds"][:], z5_file["ds"][:])
    assert blocks_in(file_["ds"]._path) == blocks_in(z5_path / "ds")

    attrs = attrs_in(file_["ds"]._path)
    z5_attrs = attrs_in(z5_path / "ds")
    for key in ("blockSize", "dimensions", "dataType", "compression"):
        assert attrs[key] == z5_attrs[key]


def test_vs_z5_hash(file_, z5_file):
    """Check identical block contents to z5"""
    z5_path = Path(z5_file.path)
    shape = (10, 20)
    data = np.arange(np.product(shape)).reshape(shape)

    file_.create_dataset("ds", data=data, chunks=(6, 7), compression="raw")
    z5_file.create_dataset("ds", data=data, chunks=(6, 7), compression="raw")

    assert blocks_hash(file_._path) == blocks_hash(z5_path)


def test_compression_opts(file_, compression_name_opt):
    compression, opt = compression_name_opt

    shape = (10, 20)
    data = np.arange(np.product(shape)).reshape(shape)

    ds = file_.create_dataset(
        "ds", data=data, chunks=(10, 10), compression=compression, compression_opts=opt
    )

    compression_dict = ds.attrs._read_attributes()["compression"]
    assert compression_dict.pop("type") == compression
    if opt is not None:
        assert list(compression_dict.values()).pop() == opt



from copy import deepcopy

import pytest
import numpy as np

import pyn5

try:
    import h5py
except ImportError:
    h5py = None

from .common import z5py

DS_SIZE = (10, 10, 10)
BLOCKSIZE = (2, 2, 2)

INT_DTYPES = ["UINT8", "UINT16", "UINT32", "UINT64", "INT8", "INT16", "INT32", "INT64"]

FLOAT_DTYPES = ["FLOAT32", "FLOAT64"]


@pytest.fixture
def random():
    return np.random.RandomState(1991)


@pytest.fixture(params=INT_DTYPES + FLOAT_DTYPES)
def ds_dtype(request, tmp_path):
    dtype = request.param
    n5_path = str(tmp_path / "test.n5")
    ds_name = "ds" + dtype

    pyn5.create_dataset(n5_path, ds_name, DS_SIZE, BLOCKSIZE, dtype)
    yield pyn5.open(n5_path, ds_name, dtype, False), np.dtype(dtype.lower())


@pytest.fixture(params=[
    {"type": "raw"},
    {"type": "bzip2", "blockSize": 5},
    {"type": "gzip", "level": 5},
    # {"type": "lz4", "blockSize": 32768},
    # {"type": "xz", "preset": 3},
], ids=lambda d: d.get("type", "raw"))
def compression_dict(request):
    yield deepcopy(request.param)


@pytest.fixture
def compression_name_opt(compression_dict):
    name = compression_dict.pop("type")
    arg = list(compression_dict.values()).pop() if compression_dict else None
    yield name, arg


@pytest.fixture
def file_(tmp_path):
    f = pyn5.File(tmp_path / "test.n5")
    yield f


@pytest.fixture
def h5_file(tmp_path):
    if not h5py:
        pytest.skip("h5py not installed")

    with h5py.File(tmp_path / "test.hdf5") as f:
        yield f


@pytest.fixture
def z5_file(tmp_path):
    if not z5py:
        pytest.skip("z5py not installed")

    yield z5py.N5File(tmp_path / "test_z5.n5")



# -*- coding: utf-8 -*-

"""Top-level package for pyn5."""

__author__ = """William Hunter Patton"""
__email__ = "pattonw@hhmi.org"
__version__ = "1.0.0"

from h5py_like import Mode

from .python_wrappers import open, read, write
from .pyn5 import (
    DatasetUINT8,
    DatasetUINT16,
    DatasetUINT32,
    DatasetUINT64,
    DatasetINT8,
    DatasetINT16,
    DatasetINT32,
    DatasetINT64,
    DatasetFLOAT32,
    DatasetFLOAT64,
    create_dataset,
)
from .attributes import AttributeManager
from .dataset import Dataset
from .file_group import File, Group
from .common import CompressionType

__all__ = [
    "open",
    "read",
    "write",
    "create_dataset",
    "DatasetUINT8",
    "DatasetUINT16",
    "DatasetUINT32",
    "DatasetUINT64",
    "DatasetINT8",
    "DatasetINT16",
    "DatasetINT32",
    "DatasetINT64",
    "DatasetFLOAT32",
    "DatasetFLOAT64",
    "File",
    "Group",
    "Dataset",
    "AttributeManager",
    "CompressionType",
    "Mode",
]



import json
import hashlib
from typing import NamedTuple, Optional, Tuple

try:
    import z5py
except ImportError:
    z5py = None


def blocks_in(dpath):
    return {
        str(path.relative_to(dpath))
        for path in dpath.glob('**/*')
        if path.suffix != ".json"
    }


def iter_block_paths(dpath):
    for fpath in sorted(
        fpath for fpath in dpath.glob('**/*')
        if fpath.is_file() and fpath.suffix != ".json"
    ):
        yield fpath


def blocks_hash(dpath):
    md5 = hashlib.md5()
    for fpath in iter_block_paths(dpath):
        md5.update(str(fpath.relative_to(dpath)).encode())
        md5.update(fpath.read_bytes())

    return md5.hexdigest()


def attrs_in(dpath):
    return json.loads((dpath / "attributes.json").read_text())


class BlockContents(NamedTuple):
    mode: int
    ndim: int
    shape: Tuple[int, ...]
    num_elem: Optional[int]
    compressed_data: bytes

    @classmethod
    def from_block(cls, fpath):
        with open(fpath, 'rb') as f:
            mode = int.from_bytes(f.read(2), "big", signed=False)
            ndim = int.from_bytes(f.read(2), "big", signed=False)
            shape = tuple(
                int.from_bytes(f.read(4), "big", signed=False)
                for _ in range(ndim)
            )
            if mode:
                num_elem = int.from_bytes(f.read(4), "big", signed=False)
            else:
                num_elem = None
            data = f.read()

        return BlockContents(mode, ndim, shape, num_elem, data)



from copy import copy

from typing import Union, Tuple, Optional, Any

import numpy as np

from h5py_like import DatasetBase, AttributeManagerBase, mutation
from h5py_like.shape_utils import thread_read_fn, thread_write_fn
from pyn5.attributes import AttributeManager
from .pyn5 import (
    DatasetUINT8,
    DatasetUINT16,
    DatasetUINT32,
    DatasetUINT64,
    DatasetINT8,
    DatasetINT16,
    DatasetINT32,
    DatasetINT64,
    DatasetFLOAT32,
    DatasetFLOAT64,
)


dataset_types = {
    np.dtype("uint8"): DatasetUINT8,
    np.dtype("uint16"): DatasetUINT16,
    np.dtype("uint32"): DatasetUINT32,
    np.dtype("uint64"): DatasetUINT64,
    np.dtype("int8"): DatasetINT8,
    np.dtype("int16"): DatasetINT16,
    np.dtype("int32"): DatasetINT32,
    np.dtype("int64"): DatasetINT64,
    np.dtype("float32"): DatasetFLOAT32,
    np.dtype("float64"): DatasetFLOAT64,
}


class Dataset(DatasetBase):
    threads = None

    def __init__(self, basename: str, parent: "Group"):  # noqa would need circular imports
        """

        :param basename: basename of the dataset
        :param parent: group to which the dataset belongs
        """
        super().__init__(basename, parent)
        self._path = self.parent._path / self.basename
        self._attrs = AttributeManager.from_container(self)
        self.threads = copy(self.threads)

        attrs = self._attrs._read_attributes()

        try:
            self._shape = tuple(attrs["dimensions"][::-1])
            self._dtype = np.dtype(attrs["dataType"].lower())
            self._chunks = tuple(attrs["blockSize"][::-1])
        except KeyError:
            raise ValueError(
                f"Not a dataset (missing metadata key): " + str(self._path)
            )

        self._impl = dataset_types[self.dtype](
            str(self.file._path),
            self.name[1:],
            True,  # raise error if dataset does not exist on disk
        )

    @property
    def dims(self):
        raise NotImplementedError()

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @property
    def maxshape(self) -> Tuple[int, ...]:
        return self.shape

    @property
    def fillvalue(self) -> Any:
        return 0

    @property
    def chunks(self) -> Optional[Tuple[int, ...]]:
        return self._chunks

    def resize(self, size: Union[int, Tuple[int, ...]], axis: Optional[int] = None):
        raise NotImplementedError()

    def __getitem__(self, args) -> np.ndarray:
        def inner_fn(translation, dimensions):
            return self._impl.read_ndarray(
                translation[::-1], dimensions[::-1]
            ).transpose()

        if self.threads:
            def fn(translation, dimensions):
                return thread_read_fn(
                    translation,
                    dimensions,
                    self.chunks,
                    self.shape,
                    inner_fn,
                    self.threads
                )
        else:
            fn = inner_fn

        return self._getitem(args, fn, self._astype)

    @mutation
    def __setitem__(self, args, val):
        def inner_fn(offset, arr):
            return self._impl.write_ndarray(
                offset[::-1], arr.transpose(), self.fillvalue
            )

        if self.threads:
            def fn(offset, arr):
                return thread_write_fn(
                    offset,
                    arr,
                    self.chunks,
                    self.shape,
                    inner_fn,
                    self.threads
                )
        else:
            fn = inner_fn

        return self._setitem(args, val, fn)

    @property
    def attrs(self) -> AttributeManagerBase:
        return self._attrs



import json

import shutil
import warnings
from pathlib import Path
from typing import Iterator

import numpy as np

from h5py_like import GroupBase, FileMixin, Mode, mutation
from h5py_like.common import Name
from h5py_like.base import H5ObjectLike
from h5py_like.shape_utils import guess_chunks

from pyn5 import Dataset
from .common import compression_args
from .attributes import AttributeManager
from .pyn5 import create_dataset

N5_VERSION = "2.0.2"
N5_VERSION_INFO = tuple(int(i) for i in N5_VERSION.split('.'))


class Group(GroupBase):
    def __init__(self, basename: str, parent: "Group"):
        """

        :param basename: basename of the group
        :param parent: group to which the group belongs
        """
        super().__init__(basename, parent)
        self._path = self.parent._path / self.basename

        self._attrs = AttributeManager.from_container(self)

    def _create_child_group(self, name) -> GroupBase:
        dpath = self._path / name

        try:
            obj = self._get_child(name)
        except KeyError:
            pass
        else:
            if isinstance(obj, Group):
                raise FileExistsError(f"Group already exists at {dpath}")
            else:
                raise TypeError(f"Dataset found at {dpath}")

        dpath.mkdir()
        return Group(name, self)

    def _create_child_dataset(
        self, name, shape=None, dtype=None, data=None, chunks=None,
        compression=None, compression_opts=None, **kwds
    ):
        for key in kwds:
            warnings.warn(
                f"pyn5 does not implement '{key}' argument for create_dataset; it will be ignored"
            )

        if data is not None:
            data = np.asarray(data, dtype=dtype)
            dtype = data.dtype
            shape = data.shape

        dtype = np.dtype(dtype)

        if chunks is None:
            warnings.warn(
                "chunks not set: entire dataset will be a single chunk. "
                "This will be slow and inefficient. "
                "Set chunks=True to guess reasonable chunk sizes."
            )
            chunks = shape
        elif chunks is True:
            chunks = guess_chunks(shape, dtype.itemsize)

        dpath = self._path / name

        try:
            obj = self._get_child(name)
        except KeyError:
            pass
        else:
            if isinstance(obj, Dataset):
                raise TypeError(f"Dataset already exists at {dpath}")
            elif isinstance(obj, Group):
                raise FileExistsError(f"Group found at {dpath}")

        if compression:
            try:
                opt_name = compression_args[compression]
            except KeyError:
                raise ValueError(
                    f"Unknown compression type '{compression}': "
                    f"use one of {sorted(compression_args)}"
                )

            compression_dict = {"type": str(compression)}
            if compression_opts is not None:
                compression_dict[opt_name] = compression_opts

            compression_str = json.dumps(compression_dict)
        else:
            compression_str = None

        file_path = str(self.file.filename)
        create_dataset(
            file_path,
            str(Name(self.name) / name)[1:],
            list(shape)[::-1],
            list(chunks)[::-1],
            dtype.name.upper(),
            compression_str,
        )

        ds = Dataset(name, self)
        if data is not None:
            ds[...] = data
        return ds

    def _get_child(self, name) -> H5ObjectLike:
        dpath = self._path / name
        if not dpath.is_dir():
            raise KeyError()
        attrs = AttributeManager(dpath)
        if attrs._is_dataset():
            return Dataset(name, self)
        else:
            return Group(name, self)

    @mutation
    def __setitem__(self, name, obj):
        """Not implemented"""
        raise NotImplementedError()

    def copy(
        self,
        source,
        dest,
        name=None,
        shallow=False,
        expand_soft=False,
        expand_external=False,
        expand_refs=False,
        without_attrs=False,
    ):
        """Not implemented"""
        raise NotImplementedError()

    @property
    def attrs(self) -> AttributeManager:
        return self._attrs

    @mutation
    def __delitem__(self, v) -> None:
        shutil.rmtree(self[v]._path)

    def __len__(self) -> int:
        counter = 0
        for _ in self:
            counter += 1
        return counter

    def __iter__(self) -> Iterator:
        for path in self._path.iterdir():
            if path.is_dir():
                yield path.name


class File(FileMixin, Group):
    def __init__(self, name, mode=Mode.READ_WRITE_CREATE):
        super().__init__(name, mode)
        self._require_dir(self.filename)
        self._path = self.filename
        self._attrs = AttributeManager.from_container(self)

    @mutation
    def __setitem__(self, name, obj):
        """Not implemented"""
        raise NotImplementedError()

    def copy(
        self,
        source,
        dest,
        name=None,
        shallow=False,
        expand_soft=False,
        expand_external=False,
        expand_refs=False,
        without_attrs=False,
    ):
        """Not implemented"""
        raise NotImplementedError()

    def _require_dir(self, dpath: Path):
        if dpath.is_file():
            raise FileExistsError("File found at desired location of directory")
        created = False
        if dpath.is_dir():
            if self.mode == Mode.CREATE:
                raise FileExistsError()
            elif self.mode == Mode.CREATE_TRUNCATE:
                shutil.rmtree(dpath)
                dpath.mkdir()
                created = True
        else:
            if self.mode in (Mode.READ_ONLY, Mode.READ_WRITE):
                raise FileNotFoundError()
            else:
                dpath.mkdir(parents=True)
                created = True

        attrs = AttributeManager(dpath, self.mode)
        if created:
            attrs["n5"] = N5_VERSION
        else:
            version = attrs.get("n5")
            if not version:
                raise ValueError(f"No N5 version found in {attrs._path}")

            version_info = tuple(int(i) for i in version.split('.'))

            if version_info[0] != N5_VERSION_INFO[0]:
                raise ValueError(f"Expected N5 version '{N5_VERSION}', got {version}")
            elif version_info[1] != N5_VERSION_INFO[1]:
                warnings.warn(f"Expected N5 version '{N5_VERSION}', got {version};"
                              f" trying to open anyway")

        return created



import errno
import json
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps

from pathlib import Path

from typing import Iterator, Any, Dict

import numpy as np

from h5py_like import AttributeManagerBase, mutation, Mode
from h5py_like.base import H5ObjectLike


class NumpyEncoder(json.JSONEncoder):
    """JSON encoder object which converts numpy arrays to lists."""

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        return super().default(obj)


def restrict_metadata(fn):
    """Decorator for AttributeManager methods which prevents mutation of N5 metadata"""
    @wraps(fn)
    def wrapped(obj: "AttributeManager", key, *args, **kwargs):
        if obj._is_dataset() and key in obj._dataset_keys:
            raise RuntimeError(f"N5 metadata (key '{key}') cannot be mutated")
        return fn(obj, key, *args, **kwargs)
    return mutation(wrapped)


class AttributeManager(AttributeManagerBase):
    """Object which reads and writes group attributes as JSON.

    The ``_encoder`` member variable (default ``NumpyEncoder``) can be set on
    the class or the instance to change how attributes are serialised.

    The ``_dump_kwargs`` member variable is passed as kwargs to ``json.dump`` on write.
    By default, it is an empty dict.
    New instances make a deep copy of the class variable.
    """

    _dataset_keys = {"dimensions", "blockSize", "dataType", "compression"}
    _encoder = NumpyEncoder
    _dump_kwargs = dict()

    def __init__(self, dpath: Path, mode=Mode.default()):
        """
        :param dpath: Path of the directory in which the attributes file resides.
        :param mode: Mode
        """
        self._path = Path(dpath) / "attributes.json"
        self._dump_kwargs = deepcopy(self._dump_kwargs)
        self._has_dataset_keys_ = None
        super().__init__(mode)

    @classmethod
    def from_container(cls, container: H5ObjectLike) -> "AttributeManager":
        """
        Create AttributeManager for a File, Group or Dataset.

        :param container: File, Group or Dataset to which the attributes belong.
        :return: AttributeManager instance
        """
        return cls(container._path, container.mode)

    @restrict_metadata
    def __setitem__(self, k, v) -> None:
        with self._open_attributes(True) as attrs:
            attrs[k] = v

    @restrict_metadata
    def __delitem__(self, v) -> None:
        with self._open_attributes(True) as attrs:
            del attrs[v]

    def __getitem__(self, k):
        with self._open_attributes() as attrs:
            return attrs[k]

    def __len__(self) -> int:
        with self._open_attributes() as attrs:
            return len(attrs)

    def __iter__(self) -> Iterator:
        yield from self.keys()

    def keys(self):
        with self._open_attributes() as attrs:
            return attrs.keys()

    def values(self):
        """Mutations are not written back to the attributes file"""
        with self._open_attributes() as attrs:
            return attrs.values()

    def items(self):
        """Mutations are not written back to the attributes file"""
        with self._open_attributes() as attrs:
            return attrs.items()

    def __contains__(self, item):
        with self._open_attributes() as attrs:
            return item in attrs

    def _is_dataset(self) -> bool:
        if self._has_dataset_keys_ is None:
            try:
                with open(self._path, "r") as f:
                    self._has_dataset_keys_ = self._dataset_keys.issubset(json.load(f))
            except ValueError:
                self._has_dataset_keys_ = False
            except IOError as e:
                if e.errno == errno.ENOENT:
                    self._has_dataset_keys_ = False
                else:
                    raise
        return self._has_dataset_keys_

    @contextmanager
    def _open_attributes(self, write: bool = False) -> Dict[str, Any]:
        """Return attributes as a context manager.

        N5 metadata keys are stripped from the dict.

        :param write: Whether to write changes to the attributes dict.
        :return: attributes as a dict
        """
        attributes = self._read_attributes()

        if self._is_dataset():
            hidden_attrs = {k: attributes.pop(k) for k in self._dataset_keys}
        else:
            hidden_attrs = dict()

        yield attributes
        if write:
            hidden_attrs.update(attributes)
            self._write_attributes(hidden_attrs)

    def _read_attributes(self):
        """Return attributes or an empty dict if they do not exist"""
        try:
            with open(self._path, "r") as f:
                attributes = json.load(f)
        except ValueError:
            attributes = {}
        except IOError as e:
            if e.errno == errno.ENOENT:
                attributes = {}
            else:
                raise

        return attributes

    def _write_attributes(self, attrs):
        """Write dict to attributes file, using AttributeManager's encoder and kwargs."""
        with open(self._path, "w") as f:
            json.dump(attrs, f, cls=self._encoder, **self._dump_kwargs)



import enum


class StrEnum(str, enum.Enum):
    def __new__(cls, *args):
        for arg in args:
            if not isinstance(arg, (str, enum.auto)):
                raise TypeError(
                    "Values of StrEnums must be strings: {} is a {}".format(
                        repr(arg), type(arg)
                    )
                )
        return super().__new__(cls, *args)

    def __str__(self):
        return self.value

    # pylint: disable=no-self-argument
    # The first argument to this function is documented to be the name of the
    # enum member, not `self`:
    # https://docs.python.org/3.6/library/enum.html#using-automatic-values
    def _generate_next_value_(name, *_):
        return name


class CompressionType(StrEnum):
    RAW = "raw"
    BZIP2 = "bzip2"
    GZIP = "gzip"
    # LZ4 = "lz4"
    # XZ = "xz"


compression_args = {
    CompressionType.RAW: None,
    CompressionType.BZIP2: "blockSize",
    CompressionType.GZIP: "level",
    # CompressionType.LZ4: "blockSize",
    # CompressionType.XZ: "preset",
}
